r"""
Loads MPT files recorded with the EC-Lab software from BioLogic for BioLogic potentiostats.

EXAMPLES:

The file can be loaded with the ECLabLoader::

    >>> from io import StringIO
    >>> file = StringIO('''EC-Lab ASCII FILE
    ... Nb header lines : 6
    ...
    ... Device metadata : some metadata
    ...
    ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
    ... 2\t0\t0.1\t0\t0
    ... 2\t1\t1.4\t5\t1
    ... ''')
    >>> eclab_csv = ECLabLoader(file)
    >>> eclab_csv.df
       mode  time/s  Ewe/V  <I>/mA  control/V
    0     2       0    0.1       0          0
    1     2       1    1.4       5          1

The file can also be loaded from the base loader::

    >>> from io import StringIO
    >>> file = StringIO('''EC-Lab ASCII FILE
    ... Nb header lines : 6
    ...
    ... Device metadata : some metadata
    ...
    ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
    ... 2\t0\t0.1\t0\t0
    ... 2\t1\t1.4\t5\t1
    ... ''')
    >>> from echemdbconverters.baseloader import BaseLoader
    >>> csv = BaseLoader.create('eclab')(file)
    >>> csv.df
       mode  time/s  Ewe/V  <I>/mA  control/V
    0     2       0    0.1       0          0
    1     2       1    1.4       5          1

    >>> csv.header.readlines()
    ['EC-Lab ASCII FILE\n', 'Nb header lines : 6\n', '\n', 'Device metadata : some metadata\n', '\n']

    >>> csv.column_header_names
    ['mode', 'time/s', 'Ewe/V', '<I>/mA', 'control/V']

"""

# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2025 Albert Engstfeld
#        Copyright (C) 2022 Johannes Hermann
#        Copyright (C) 2022 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************


from echemdbconverters.baseloader import BaseLoader


class ECLabLoader(BaseLoader):
    r"""
    Loads BioLogic EC-Lab MPT files.

    EXAMPLES::

        >>> from io import StringIO
        >>> file = StringIO('''EC-Lab ASCII FILE
        ... Nb header lines : 6
        ...
        ... Device metadata : some metadata
        ...
        ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
        ... 2\t0\t0.1\t0\t0
        ... 2\t1\t1.4\t5\t1
        ... ''')
        >>> from echemdbconverters.baseloader import BaseLoader
        >>> csv = BaseLoader.create('eclab')(file)
        >>> csv.df
           mode  time/s  Ewe/V  <I>/mA  control/V
        0     2       0    0.1       0          0
        1     2       1    1.4       5          1

        >>> csv.header.readlines()
        ['EC-Lab ASCII FILE\n', 'Nb header lines : 6\n', '\n', 'Device metadata : some metadata\n', '\n']

        >>> csv.column_header_names
        ['mode', 'time/s', 'Ewe/V', '<I>/mA', 'control/V']

    """

    @property
    def header_lines(self):
        r"""
        The number of header lines of an EC-Lab MPT file without column names.
        The number is provided in the header of the MPT file, which contains, however,
        also the line with the data column names.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO('''EC-Lab ASCII FILE
            ... Nb header lines : 6
            ...
            ... Device metadata : some metadata
            ...
            ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
            ... 2\t0\t0.1\t0\t0
            ... 2\t1\t1.4\t5\t1
            ... ''')
            >>> from echemdbconverters.baseloader import BaseLoader
            >>> csv = BaseLoader.create('eclab')(file)
            >>> csv.header_lines
            5

        """
        matches = []

        for line in self.file.readlines():
            import re

            match = re.findall(
                r"(?P<headerlines>Nb header lines) *\: *(?P<value>-?\d+\.?\d*)",
                str(line),
                re.IGNORECASE,
            )

            if match:
                matches.append(match)

        return int(matches[0][0][1]) - 1



r"""
Helpers for click CLI testing.

Click's own CliRunner is quite cumbersome to work with in some simple test
scenarios so we wrap it in more convenient ways here.

The code has been inherited from echemdb-converters and adapted for echemdb-converters.

"""

# *********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2021-2025 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# *********************************************************************

import pytest


def invoke(command, *args):
    r"""
    Invoke the click ``command`` with the given list of string arguments.

    >>> import click
    >>> @click.command()
    ... def hello(): print("Hello World")
    >>> invoke(hello)
    Hello World

    >>> @click.command()
    ... def fails(): raise Exception("expected error")
    >>> invoke(fails)
    Traceback (most recent call last):
    ...
    Exception: expected error

    """
    from click.testing import CliRunner

    invocation = CliRunner().invoke(command, args, catch_exceptions=False)
    output = invocation.output.strip()
    if output:
        print(output)


class TemporaryData:
    r"""
    Provides a temporary directory with test files.

    EXAMPLES::

        >>> import os
        >>> with TemporaryData("unit.*") as directory:
        ...     'unit.csv' in os.listdir(directory)
        True

    """

    def __init__(self, *patterns):
        self._patterns = patterns
        self._tmpdir = None

    def __enter__(self):
        import tempfile

        self._tmpdir = tempfile.TemporaryDirectory()

        try:
            import glob
            import os
            import os.path
            import shutil

            import echemdbconverters

            cwd = os.getcwd()
            os.chdir(
                os.path.join(
                    os.path.dirname(echemdbconverters.__file__), "..", "test", "data"
                )
            )
            try:
                for pattern in self._patterns:
                    for filename in glob.glob(pattern):
                        shutil.copy(filename, self._tmpdir.name)

                return self._tmpdir.name
            finally:
                os.chdir(cwd)
            self._tmpdir.__enter__()
        except Exception:
            self._tmpdir.cleanup()
            raise

    def __exit__(self, *args):
        self._tmpdir.__exit__(*args)


@pytest.mark.parametrize(
    "name,args",
    [
        (  # "Standard" CSV with a single column header line.
            "default",
            ["csv", "default.csv"],
        ),
        (  # "Standard" CSV for which the units to the columns are included in an additional metadata file.
            "unit",
            ["csv", "unit.csv", "--metadata", "unit.csv.metadata"],
        ),
        (  # Biologic EClab file for cyclic voltammetry (using `,` as decimal separator)
            "eclab_cv",
            [
                "csv",
                "eclab_cv.mpt",
                "--metadata",
                "eclab_cv.mpt.metadata",
                "--device",
                "eclab",
            ],
        ),
        (  # Biologic EClab file for chronoamperrometry or chronopotentiometry (using `,` as decimal separator)
            "eclab_ca",
            [
                "csv",
                "eclab_ca.mpt",
                "--metadata",
                "eclab_ca.mpt.metadata",
                "--device",
                "eclab",
            ],
        ),
        (  # Gamry Instruments Framework software file for cyclic voltammetry
            "gamry_cv",
            [
                "csv",
                "gamry_cv.DTA",
                "--metadata",
                "gamry_cv.DTA.metadata",
                "--device",
                "gamry",
            ],
        ),
    ],
)
def test_csv(name, args):
    r"""
    Test that the csv command from the command line interface works correctly.

    This function is executed by pytest and checks that "csv" produces JSON and
    CSV files that match expected outputs.
    """
    import os

    cwd = os.getcwd()
    with TemporaryData(f"{name}.*") as workdir:
        os.chdir(workdir)
        try:
            from echemdbconverters.entrypoint import cli

            invoke(cli, *args, "--outdir", "outdir")

            import json

            with open(f"outdir/{name}.json", encoding="ASCII") as actual:
                with open(f"{name}.json.expected", encoding="ASCII") as expected:
                    assert json.load(actual) == json.load(expected)

            import pandas
            import pandas.testing

            pandas.testing.assert_frame_equal(
                pandas.read_csv(f"outdir/{name}.csv"),
                pandas.read_csv(f"{name}.csv.expected"),
            )
        finally:
            os.chdir(cwd)



r"""
Loads DAT files recorded with the Gamry Instruments Framework software
from Gamry for Gamry potentiostats.

EXAMPLES:

The file can be loaded with the GamryLoader::

    >>> from io import StringIO
    >>> file = StringIO('''EXPLAIN
    ... TAG\tCV
    ... TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier
    ... CURVE\tTABLE\t3597
    ... \tPt\tT\tVf\tIm\tVu\tSig\tAch\tIERange\tOver\tCycle\tTemp
    ... \t#\ts\tV vs. Ref.\tA\tV\tV\tV\t#\tbits\t#\tdeg C
    ... \t0\t0,06\t2,00054E-001\t1,72821E-005\t0,00000E+000\t2,00000E-001\t6,45222E-004\t9\t..........a\t0\t-327,75
    ... \t1\t0,12\t1,97170E-001\t1,04547E-005\t0,00000E+000\t1,97000E-001\t-1,17889E-003\t9\t..........a\t0\t-327,75
    ... ''')
    >>> from echemdbconverters.gamryloader import GamryLoader
    >>> csv = GamryLoader(file)
    >>> csv.df # doctest: +NORMALIZE_WHITESPACE
       Pt / #  T / s  Vf / V vs. Ref.  ...  Over / bits  Cycle / #  Temp / deg C
    0       0   0.06         0.200054  ...  ..........a          0       -327.75
    1       1   0.12         0.197170  ...  ..........a          0       -327.75
    ...

The file can also be loaded from the base loader::

    >>> from io import StringIO
    >>> file = StringIO('''EXPLAIN
    ... TAG\tCV
    ... TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier
    ... CURVE\tTABLE\t3597
    ... \tPt\tT\tVf\tIm\tVu\tSig\tAch\tIERange\tOver\tCycle\tTemp
    ... \t#\ts\tV vs. Ref.\tA\tV\tV\tV\t#\tbits\t#\tdeg C
    ... \t0\t0,06\t2,00054E-001\t1,72821E-005\t0,00000E+000\t2,00000E-001\t6,45222E-004\t9\t..........a\t0\t-327,75
    ... \t1\t0,12\t1,97170E-001\t1,04547E-005\t0,00000E+000\t1,97000E-001\t-1,17889E-003\t9\t..........a\t0\t-327,75
    ... ''')
    >>> from echemdbconverters.baseloader import BaseLoader
    >>> csv = BaseLoader.create('gamry')(file)
    >>> csv.df
       Pt / #  T / s  Vf / V vs. Ref.  ...  Over / bits  Cycle / #  Temp / deg C
    0       0   0.06         0.200054  ...  ..........a          0       -327.75
    1       1   0.12         0.197170  ...  ..........a          0       -327.75
    ...

    >>> csv.header.readlines()
    ['EXPLAIN\n', 'TAG\tCV\n', 'TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier\n', 'CURVE\tTABLE\t3597\n']

    >>> csv.column_header_names
    ['Pt / #', 'T / s', 'Vf / V vs. Ref.', 'Im / A', 'Vu / V', 'Sig / V', 'Ach / V', 'IERange / #', 'Over / bits', 'Cycle / #', 'Temp / deg C']

"""

# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2025 Albert Engstfeld
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************


from echemdbconverters.baseloader import BaseLoader


class GamryLoader(BaseLoader):
    r"""
    Loads Gamry Instruments Framework DAT files.

    EXAMPLES::

        >>> from io import StringIO
        >>> file = StringIO('''EXPLAIN
        ... TAG\tCV
        ... TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier
        ... CURVE\tTABLE\t3597
        ... \tPt\tT\tVf\tIm\tVu\tSig\tAch\tIERange\tOver\tCycle\tTemp
        ... \t#\ts\tV vs. Ref.\tA\tV\tV\tV\t#\tbits\t#\tdeg C
        ... \t0\t0,06\t2,00054E-001\t1,72821E-005\t0,00000E+000\t2,00000E-001\t6,45222E-004\t9\t..........a\t0\t-327,75
        ... \t1\t0,12\t1,97170E-001\t1,04547E-005\t0,00000E+000\t1,97000E-001\t-1,17889E-003\t9\t..........a\t0\t-327,75
        ... ''')
        >>> from echemdbconverters.baseloader import BaseLoader
        >>> csv = BaseLoader.create('gamry')(file)
        >>> csv.df
           Pt / #  T / s  Vf / V vs. Ref.  ...  Over / bits  Cycle / #  Temp / deg C
        0       0   0.06         0.200054  ...  ..........a          0       -327.75
        1       1   0.12         0.197170  ...  ..........a          0       -327.75
        ...

        >>> csv.header.readlines()
        ['EXPLAIN\n', 'TAG\tCV\n', 'TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier\n', 'CURVE\tTABLE\t3597\n']

        >>> csv.column_header_names
        ['Pt / #', 'T / s', 'Vf / V vs. Ref.', 'Im / A', 'Vu / V', 'Sig / V', 'Ach / V', 'IERange / #', 'Over / bits', 'Cycle / #', 'Temp / deg C']

    """

    def _included_files(self):
        r"""
        Some Gamry files contain several files, e.g., cycles of a CV which are split
        by lines containing ``CURVE  TABLE   <SomeNumber>``.

        We search for all those instances to construct a common df with meth df.

        """
        return NotImplementedError

    @property
    def header_lines(self):
        r"""
        The number of header lines of an EC-Lab MPT file without column names.
        The number is provided in the header of the MPT file, which contains, however,
        also the line with the data column names.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO('''EXPLAIN
            ... TAG\tCV
            ... TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier
            ... CURVE\tTABLE\t3597
            ... \tPt\tT\tVf\tIm\tVu\tSig\tAch\tIERange\tOver\tCycle\tTemp
            ... \t#\ts\tV vs. Ref.\tA\tV\tV\tV\t#\tbits\t#\tdeg C
            ... \t0\t0,06\t2,00054E-001\t1,72821E-005\t0,00000E+000\t2,00000E-001\t6,45222E-004\t9\t..........a\t0\t-327,75
            ... \t1\t0,12\t1,97170E-001\t1,04547E-005\t0,00000E+000\t1,97000E-001\t-1,17889E-003\t9\t..........a\t0\t-327,75
            ... ''')
            >>> from echemdbconverters.baseloader import BaseLoader
            >>> csv = BaseLoader.create('gamry')(file)
            >>> csv.header_lines
            4

        """

        import re

        expression = re.compile(r"CURVE\tTABLE\t(\d+)")

        for idx, line in enumerate(self.file.readlines()):
            if expression.match(line):
                return idx + 1

        raise KeyError("Could not find a line containing `Curve Label` in the file.")

    @property
    def column_header_lines(self):
        r"""The number of lines containing descriptive information
        on the columns of Gamry DAT files is 2.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO('''EXPLAIN
            ... TAG\tCV
            ... TITLE\tLABEL\tCyclic Voltammetry\tTest &Identifier
            ... CURVE\tTABLE\t3597
            ... \tPt\tT\tVf\tIm\tVu\tSig\tAch\tIERange\tOver\tCycle\tTemp
            ... \t#\ts\tV vs. Ref.\tA\tV\tV\tV\t#\tbits\t#\tdeg C
            ... \t0\t0,06\t2,00054E-001\t1,72821E-005\t0,00000E+000\t2,00000E-001\t6,45222E-004\t9\t..........a\t0\t-327,75
            ... \t1\t0,12\t1,97170E-001\t1,04547E-005\t0,00000E+000\t1,97000E-001\t-1,17889E-003\t9\t..........a\t0\t-327,75
            ... ''')
            >>> from echemdbconverters.baseloader import BaseLoader
            >>> csv = BaseLoader.create('gamry')(file)
            >>> csv.column_header_lines
            2

        """
        return 2



project = 'echemdb-converters'
copyright = '2022-2023, the echemdb-converters authors'
author = 'the echemdb-converters authors'

release = '0.4.0'

extensions = ["sphinx.ext.autodoc", "sphinx.ext.todo", "myst_nb"]

source_suffix = [".rst", ".md"]

templates_path = ['_templates']

exclude_patterns = ['generated', 'Thumbs.db', '.DS_Store', 'README.md', 'news', '.ipynb_checkpoints', '*.ipynb', '**/*.ipynb']

myst_enable_extensions = ["amsmath", "dollarmath"]

todo_include_todos = True

html_theme = 'sphinx_rtd_theme'

nb_execution_timeout = 90

html_static_path = []

myst_heading_anchors = 2

# a warning is raised which is treated as an error when plotly plots are rendered.
# The plots are created even when the warning is ignored.
# Warning, treated as error:
# X:\github\unitpackage\doc\index.md:82:skipping unknown output mime type: application/vnd.plotly.v1+json [mystnb.unknown_mime_type]
suppress_warnings = ["mystnb.unknown_mime_type"]

# We render some demo notebooks as part of the documentation. These notebooks try to load plotly through RequireJS (which is how Jupyter notebooks load dependencies) so we need to ship this with our Sphinx documentation.
html_js_files = ["https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"]

# Add Edit on GitHub links
html_context = {
    'display_github': True,
    'github_user': 'echemdb',
    'github_repo': 'echemdb-converters',
    'github_version': 'main/doc/',
}
